//  This file is distributed as part of the bit-babbler package.
//  Copyright 2010 - 2015,  Ron <ron@debian.org>

#ifndef _BB_SECRET_SOURCE_H
#define _BB_SECRET_SOURCE_H

#include <bit-babbler/health-monitor.h>

#include <tr1/unordered_map>
#include <unistd.h>

#if EM_PLATFORM_MSW
 // We don't actually need this (socket support) here, but the crazy windows API
 // requires winsock2.h to be included before windows.h, and ftdi.h pulls in the
 // latter (via the libusb header).
 #include <ws2tcpip.h>
#endif

#include <ftdi.h>

#if EM_PLATFORM_LINUX
 #include <linux/random.h>
 #include <sys/ioctl.h>
 #include <poll.h>
 #include <fcntl.h>
#endif

#if HAVE_LIBUDEV
// Prior to udev-150 their headers don't wrap this for C++, so do it here until
// the buggy distro releases catch up.  FC12 and CentOS 6, I'm looking at you.
extern "C" {
 #include <libudev.h>
}
#endif


#define FTDI_VENDOR_ID      0x0403
#define FTDI_PRODUCT_ID     0x6014
#define BB_VENDOR_ID        FTDI_VENDOR_ID
#define BB_PRODUCT_ID       0x7840

#define BB_WHITE_PRODUCTSTR "White RNG"
#define BB_BLACK_PRODUCTSTR "Black RNG"

// We sample on the rising edge and clock bits out MSB first by default
//#define SAMPLE_FALLING_EDGE
//#define LSB_FIRST


namespace BitB
{

    struct USBInfo
    { //{{{

        std::string     mfg;
        std::string     product;
        std::string     serial;


        USBInfo( struct usb_device *dev )
        { //{{{

            if( ! dev )
                return;

            usb_dev_handle *h = usb_open( dev );
            char            s[128];

            // In theory we could use bus->location here, but setting that
            // (from dirname) appears to be a Debian specific patch right now.
            if( ! h )
                throw Error( _("USBInfo: failed to open devnum %lu on bus %lu"),
                             StrToUL(dev->filename, 10), StrToUL(dev->bus->dirname, 10) );

            // usb_strerror() is a little useless here unfortunately, it isn't always
            // set for every reason this might fail.  And isn't thread-safe either.
            // If this is ever really expected to fail, we should probably not use it.

            if( dev->descriptor.iManufacturer )
            {
                if( usb_get_string_simple( h, dev->descriptor.iManufacturer, s, sizeof(s) ) < 0 )
                {
                    Log<0>( _("USBInfo: failed to get manufacturer descriptor "
                                        "for devnum %lu on bus %lu: %s\n"),
                                        StrToUL(dev->filename, 10),
                                        StrToUL(dev->bus->dirname, 10), usb_strerror() );

                    // The most probable reason for this to have failed is that the
                    // user attempting it does not have permission to read from the
                    // device.  In which case trying to read the other strings below
                    // is going to fail in the same way too, so just short-circuit
                    // that instead of creating a flood of error messages.  It is
                    // possible this might fail due to malformed or corrupt data too
                    // but in that case you've probably got bigger things to worry
                    // about than us not extracting the other fields here. The lsusb
                    // utility can still show you that.
                    goto hell;
                }
                mfg = s;
            }

            if( dev->descriptor.iProduct )
            {
                if( usb_get_string_simple( h, dev->descriptor.iProduct, s, sizeof(s) ) < 0 )
                {
                    Log<0>( _("USBInfo: failed to get product descriptor "
                                        "for devnum %lu on bus %lu: %s\n"),
                                        StrToUL(dev->filename, 10),
                                        StrToUL(dev->bus->dirname, 10), usb_strerror() );
                    goto hell;
                }
                product = s;
            }

            if( dev->descriptor.iSerialNumber )
            {
                if( usb_get_string_simple( h, dev->descriptor.iSerialNumber, s, sizeof(s) ) < 0 )
                {
                    Log<0>( _("USBInfo: failed to get serial descriptor "
                                        "for devnum %lu on bus %lu: %s\n"),
                                        StrToUL(dev->filename, 10),
                                        StrToUL(dev->bus->dirname, 10), usb_strerror() );
                    goto hell;
                }
                serial = s;
            }

        hell:
            usb_close( h );

        } //}}}

    }; //}}}


    struct DeviceInfo
    { //{{{

        typedef std::vector< DeviceInfo >   Vector;


        std::string     mfg;
        std::string     product;
        std::string     serial;
        unsigned        busnum;
        unsigned        devnum;
        std::string     devport;


        DeviceInfo() {}

        DeviceInfo( const std::string  &m,
                    const std::string  &p,
                    const std::string  &s,
                    unsigned            bn,
                    unsigned            dn,
                    const std::string  &dp = std::string() )
            : mfg( m )
            , product( p )
            , serial( s )
            , busnum( bn )
            , devnum( dn )
            , devport( dp )
        {}


        std::string DevicePort() const
        {
            return stringprintf( "%u-", busnum ) + devport;
        }

        std::string IDStr() const
        {
            if( devport.empty() )
            {
                if( serial.empty() )
                    return stringprintf( "Serial '', bus %u, device %u", busnum, devnum );

                return "Serial '" + serial + "'";
            }

            return stringprintf( "Serial '%s', port %u-%s",
                                 serial.c_str(), busnum, devport.c_str() );
        }

        std::string ProductStr() const
        {
            return stringprintf( "Serial '%s', Mfg '%s', Product '%s'",
                                 serial.c_str(), mfg.c_str(), product.c_str() );
        }

        std::string VerboseStr() const
        {
            // If not using udev we can't know the physical port the device is
            // attached to without walking the sysfs tree ourselves, in which
            // case, you might as well have used udev, and if that's not an
            // option, you probably don't have sysfs either ...
            if( devport.empty() )
                return stringprintf( "Serial '%s', Mfg '%s', Product '%s', "
                                     "bus %u, device %u",
                                     serial.c_str(), mfg.c_str(), product.c_str(),
                                                                busnum, devnum );

            return stringprintf( "Serial '%s', Mfg '%s', Product '%s', "
                                 "bus %u, device %u, port %s",
                                 serial.c_str(), mfg.c_str(), product.c_str(),
                                         busnum, devnum, DevicePort().c_str() );
        }

    }; //}}}


    class FTDI : public RefCounted
    { //{{{
    protected:

        static const unsigned   FTDI_READ_RETRIES = 10;

        mutable struct ftdi_context     m_ftdi;
        USBInfo                         m_usbinfo;


    private:

        // Returns true if initial FTDI communication is successful
        bool check_sync( uint8_t cmd )
        { //{{{

            uint8_t     buf[8192];
            unsigned    n = 0;

            LogMsg<3>( "FTDI::check_sync( %02x )", cmd );

            WriteCommand( OctetString( 1, cmd ) );

            do {
                int ret = ftdi_read_data( &m_ftdi, buf, sizeof(buf) );

                if( ret < 0 )
                    ThrowError( _("check_sync failed") );

                if( ret == 2 && buf[0] == 0xFA && buf[1] == cmd )
                {
                    LogMsg<3>( "have sync for 0x%x (n = %u)", cmd, n );
                    return true;
                }

                if( ret > 0 )
                {
                    LogMsg<3>( "sync returned %d bytes (n = %u)", ret, n );

                    n = 0;

                    if( opt_verbose > 3 )
                    {
                        int             nbytes = 16;
                        std::string     msg;

                        for( int i = 0; i < std::min(ret,nbytes); ++i )
                            msg += stringprintf( " %02x", buf[i] );

                        if( ret > nbytes )
                            LogMsg<4>( "%s ...", msg.c_str() );
                        else
                            LogMsg<4>( "%s", msg.c_str() );
                    }
                }

            } while( ++n < FTDI_READ_RETRIES );

            return false;

        } //}}}


    protected:

        // Returns the total number of bytes purged
        size_t purge_read()
        { //{{{

            uint8_t     buf[8192];
            size_t      count = 0;
            unsigned    n = 0;

            LogMsg<3>( "FTDI::purge_read" );

            do {
                int ret = ftdi_read_data( &m_ftdi, buf, sizeof(buf) );

                if( ret < 0 )
                    ThrowError( _("purge_read failed") );

                if( ret > 0 )
                {
                    count += ret;
                    LogMsg<3>( "purged %d / %zu (n = %u)", ret, count, n );
                    n = 0;

                    if( opt_verbose > 3 )
                    {
                        int             nbytes = 16;
                        std::string     msg;

                        for( int i = 0; i < std::min(ret,nbytes); ++i )
                            msg += stringprintf(" %02x", buf[i]);

                        if( ret > nbytes )
                            LogMsg<4>( "%s ...", msg.c_str() );
                        else
                            LogMsg<4>( "%s", msg.c_str() );
                    }
                }

            } while( ++n < FTDI_READ_RETRIES );

            return count;

        } //}}}


        // Put the chip into MPSSE mode
        bool InitMPSSE( unsigned chunksize = 4096, uint8_t latency = 1 )
        { //{{{

            // Initialise MPSSE mode (ref AN-135 4.2)
            if( ftdi_usb_reset( &m_ftdi ) )
                ThrowError( _("FTDI::InitMPSSE: reset failed") );

            purge_read();

            // This appears to be useless here, we need to actually read from it.
            //if( ftdi_usb_purge_buffers( &m_ftdi ) )
            //    ThrowError( _("FTDI::InitMPSSE: purge failed") );

            if( ftdi_read_data_set_chunksize( &m_ftdi, chunksize ) )
                ThrowError( _("FTDI::InitMPSSE: set read chunksize failed") );

            if( ftdi_set_event_char( &m_ftdi, '\0', 0 ) )
                ThrowError( _("FTDI::InitMPSSE: set event char failed") );

            if( ftdi_set_error_char( &m_ftdi, '\0', 0 ) )
                ThrowError( _("FTDI::InitMPSSE: set error char failed") );

            // Read and Write timeouts are hardcoded in libftdi to 5 sec.

            if( ftdi_set_latency_timer( &m_ftdi, latency ) )
                ThrowError( _("FTDI::InitMPSSE: set latency failed") );

            if( ftdi_setflowctrl( &m_ftdi, SIO_RTS_CTS_HS ) )
                ThrowError( _("FTDI::InitMPSSE: set flow control failed") );

            if( ftdi_set_bitmode( &m_ftdi, 0, BITMODE_RESET ) )
                ThrowError( _("FTDI::InitMPSSE: MPSSE reset failed") );

            if( ftdi_set_bitmode( &m_ftdi, 0, BITMODE_MPSSE ) )
                ThrowError( _("FTDI::InitMPSSE: set bitmode failed") );

            // Wait 50ms for all of that to settle
            usleep(50000);


            try {
                return check_sync( 0xAA ) && check_sync( 0xAB );
            }
            BB_CATCH_ALL( 0, _("FTDI::InitMPSSE: sync failed") )

            return false;

        } //}}}


    public:

        FTDI( struct usb_device *dev = NULL )
            : m_usbinfo( dev )
        { //{{{

            try {
                if( ftdi_init( &m_ftdi ) )
                    ThrowError( _("FTDI init failed") );

                m_ftdi.module_detach_mode = DONT_DETACH_SIO_MODULE;

                if( dev && ftdi_usb_open_dev( &m_ftdi, dev ) )
                    ThrowError( _("FTDI device open failed") );
            }
            catch( const std::exception& )
            {
                // We need to do this even if ftdi_init fails, since it may have
                // still allocated some memory without cleaning up after itself.
                ftdi_deinit( &m_ftdi );
                throw;
            }

        } //}}}

        ~FTDI()
        { //{{{

            LogMsg<2>( "- FTDI" );

            // We need to peek inside m_ftdi here, because _usb_async_cleanup()
            // which is called from ftdi_usb_close(), doesn't check if usb_dev
            // is valid before dereferencing it, and we may be here without ever
            // having opened a device.
            if( m_ftdi.usb_dev )
            {
                try {
                    purge_read();
                }
                BB_CATCH_ALL( 0, _("FTDI: shutdown purge failed") )

                ftdi_usb_purge_buffers( &m_ftdi );
                ftdi_set_bitmode( &m_ftdi, 0, BITMODE_RESET );
                ftdi_usb_reset( &m_ftdi );
                ftdi_usb_close( &m_ftdi );
            }

            ftdi_deinit( &m_ftdi );

        } //}}}


        const std::string &GetSerial() const
        {
            return m_usbinfo.serial;
        }


        void WriteCommand( const OctetString &cmd )
        { //{{{

            LogMsg<6>( "FTDI::WriteCommand(%s )", OctetsToHex(cmd).c_str() );

            // The ftdi_write_data prototype isn't const correct,
            // but it doesn't modify the data passed to it.
            int ret = ftdi_write_data( &m_ftdi, const_cast<uint8_t*>(cmd.data()),
                                                                     cmd.size() );
            if( ret != ssize_t(cmd.size()) )
                ThrowError( _("FTDI::WriteCommand failed") );

        } //}}}


        // Convenience methods for Logging
        //{{{

        BB_PRINTF_FORMAT(2,3)
        std::string ErrorStr( const char *format, ... ) const
        { //{{{

            va_list         arglist;
            std::string     msg( m_usbinfo.serial );

            va_start( arglist, format );
            msg.append( ": " )
               .append( vstringprintf( format, arglist ) )
               .append( ": " )
               .append( ftdi_get_error_string(&m_ftdi) );
            va_end( arglist );

            return msg;

        } //}}}

        template< int N >
        BB_PRINTF_FORMAT(2,3)
        void LogError( const char *format, ... ) const
        { //{{{

            va_list         arglist;
            std::string     msg( m_usbinfo.serial );

            va_start( arglist, format );
            msg.append( ": " )
               .append( vstringprintf( format, arglist ) )
               .append( ": " )
               .append( ftdi_get_error_string(&m_ftdi) );
            va_end( arglist );

            Log<N>( "%s\n", msg.c_str() );

        } //}}}

        BB_PRINTF_FORMAT(2,3,noreturn)
        void ThrowError( const char *format, ... ) const
        { //{{{

            va_list         arglist;
            std::string     msg( m_usbinfo.serial );

            va_start( arglist, format );
            msg.append( ": " )
               .append( vstringprintf( format, arglist ) )
               .append( ": " )
               .append( ftdi_get_error_string(&m_ftdi) );
            va_end( arglist );

            throw Error( msg );

        } //}}}


        BB_PRINTF_FORMAT(2,3)
        std::string MsgStr( const char *format, ... ) const
        { //{{{

            va_list         arglist;
            std::string     msg( m_usbinfo.serial );

            va_start( arglist, format );
            msg.append( ": " )
               .append( vstringprintf( format, arglist ) );
            va_end( arglist );

            return msg;

        } //}}}

        template< int N >
        BB_PRINTF_FORMAT(2,3)
        void LogMsg( const char *format, ... ) const
        { //{{{

            va_list         arglist;
            std::string     msg( m_usbinfo.serial );

            va_start( arglist, format );
            msg.append( ": " )
               .append( vstringprintf( format, arglist ) );
            va_end( arglist );

            Log<N>( "%s\n", msg.c_str() );

        } //}}}

        //}}}

    }; //}}}


    class BitBabbler : public FTDI
    { //{{{
    private:

        static const unsigned   FTDI_INIT_RETRIES = 20;

        unsigned        m_enable_mask;
        unsigned        m_disable_pol;
        unsigned        m_bitrate;
        unsigned        m_fold;
        unsigned        m_chunksize;
        uint8_t         m_latency;
        bool            m_no_qa;


        void init_device()
        { //{{{

            for( unsigned retries = FTDI_INIT_RETRIES; retries; --retries )
            {
                if( retries < FTDI_INIT_RETRIES )
                    LogMsg<2>("BitBabbler::init_device: retrying");

                if( ! InitMPSSE( m_chunksize, m_latency ) )
                    continue;

                OctetString     cmd;

                cmd.append( 1, 0x8A );              // Disable clock divide by 5
                cmd.append( 1, 0x97 );              // Disable adaptive clocking
                cmd.append( 1, 0x8D );              // Disable 3 phase data clock

                cmd.append( 1, 0x80 );                  // Set low data pins
               #ifdef SAMPLE_FALLING_EDGE
                cmd.append( 1, 0x01 | m_disable_pol  ); //  all outputs low, clk high.
               #else
                cmd.append( 1, 0x00 | m_disable_pol );  //  CLK, DO, CS, outputs low
               #endif                                   //  masking polarity as per config
                cmd.append( 1, 0x0B | m_enable_mask );  //  Set CLK, DO, CS, as outputs,
                                                        //  generator mask as per config

                cmd.append( 1, 0x82 );              // Set high data pins
                cmd.append( 1, 0x00 );              //  all outputs low
                cmd.append( 1, 0x00 );              //  set all pins as inputs


                unsigned    clk_div = 30000000 / m_bitrate - 1;

                cmd.append( 1, 0x86 );              // Set CLK divisor
                cmd.append( 1, clk_div & 0xFF );    // divisor low byte
                cmd.append( 1, clk_div >> 8 );      // divisor high byte

                try {
                    WriteCommand( cmd );
                }
                BB_CATCH_ALL( 0, _("BitBabbler::init_device: set clock failed"), continue; )

                // Wait another 30ms for all of this to settle too
                usleep(30000);

                // Clear the (empty) return from the WriteCommand()
                purge_read();

                return;
            }

            ThrowError( _("BitBabbler::init_device failed") );

        } //}}}


    public:

        typedef RefPtr< BitBabbler >    Handle;


        struct Options
        { //{{{

            typedef std::list< Options >    List;

            std::string     device_id;
            unsigned        device_num;
            unsigned        enable_mask;
            unsigned        disable_polarity;
            unsigned        bitrate;
            unsigned        fold;
            unsigned        group;
            bool            no_qa;

            Options()
                : device_num( 0 )
                , enable_mask( 0x0f )
                , disable_polarity( 0x00 )
                , bitrate( 0 )
                , fold( unsigned(-1) )
                , group( 0 )
                , no_qa( false )
            {}

        }; //}}}


        static unsigned RealBitrate( unsigned bitrate )
        { //{{{

            if( bitrate >= 30000000 )
                return 30000000;

            if( bitrate <= 458 )
                return 458;

            return 30000000 / (30000000 / bitrate);

        } //}}}


    private:

        unsigned choose_bitrate( const Options &opt )
        { //{{{

            if( opt.bitrate )
                return RealBitrate( opt.bitrate );

            return 2500000;

        } //}}}

        unsigned choose_folding( const Options &opt )
        { //{{{

            if( opt.fold != unsigned(-1) )
                return opt.fold;

            if( m_usbinfo.product == BB_WHITE_PRODUCTSTR )
                return 1;

            if( m_usbinfo.product == BB_BLACK_PRODUCTSTR )
                return 3;

            ThrowError( _("BitBabbler: unknown product '%s', and folding not set"),
                                                        m_usbinfo.product.c_str() );
           #if EM_PLATFORM_MSW
            // Really we'll never get here, but mingw-w64 4.9.2-21+15.4
            // appears to be too stupid to figure that out ...
            return 0;
           #endif

        } //}}}


    public:

        BitBabbler( struct usb_device *dev, const Options &options = Options() )
            : FTDI( dev )
            , m_enable_mask( ~options.enable_mask << 4 & 0xf0 )
            , m_disable_pol( options.disable_polarity << 4 & 0xf0 )
            , m_bitrate( choose_bitrate(options) )
            , m_fold( choose_folding(options) )
            , m_no_qa( options.no_qa )
        { //{{{

            if( options.bitrate == m_bitrate )
                LogMsg<2>( "+ BitBabbler( bitrate %u, fold %u, mask 0x%02x )",
                                        m_bitrate, m_fold, m_enable_mask );
            else
                LogMsg<2>( "+ BitBabbler( bitrate %u (%u), fold %u, mask 0x%02x )",
                                options.bitrate, m_bitrate, m_fold, m_enable_mask );


            // Select the chunk size to be the largest power of 2 between the
            // maximum packet size and 64kB that will take less than 250ms to
            // transfer (which is then the maximum time we'll block waiting to
            // perform an orderly exit).
            m_chunksize = std::max( m_ftdi.max_packet_size,
                          std::min( 65536u,
                                    powof2_down(m_bitrate / 32
                                                          / m_ftdi.max_packet_size
                                                          * m_ftdi.max_packet_size) ) );
            m_latency = std::max( 1u,
                        std::min( 255u,
                                  m_ftdi.max_packet_size * 8000 / m_bitrate + 1 ) );

            LogMsg<3>( "Chunk size %u, %u ms/per chunk (latency %u ms, max packet %u)",
                                m_chunksize, m_chunksize * 8000 / m_bitrate, m_latency,
                                                                m_ftdi.max_packet_size );

            // The above computes what should be our theoretical optimum latency,
            // ie. the amount of time it would take to completely fill a packet
            // of the maximum allowable size.  That in theory, should give us the
            // best throughput, since it requires the least number of transactions
            // to complete the transfer.  In practice however, it appears that the
            // throughput is worse, at any speed, if the latency is greater than 1.
            // That is quite possibly due to the way libftdi fudges the requests
            // it makes when it strips off the extra status bytes, but it could be
            // braindeath somewhere else too.  So for now, just use what works ...
            m_latency = 1;

            init_device();

        } //}}}

        ~BitBabbler()
        {
            LogMsg<2>( "- BitBabbler" );
        }


        unsigned GetBitrate() const
        {
            return m_bitrate;
        }

        unsigned GetFolding() const
        {
            return m_fold;
        }

        unsigned GetChunkSize() const
        {
            return m_chunksize;
        }

        bool NoQA() const
        {
            return m_no_qa;
        }


        size_t read( uint8_t *buf, size_t len )
        { //{{{

            OctetString     cmd;

            if( len < 1 || len > 65536 )
                throw Error( _("BitBabbler::read( %zu ): invalid length"), len );

          #ifdef LSB_FIRST

           #ifdef SAMPLE_FALLING_EDGE
            // Clock len data bytes in on -ve clock edge LSB first
            cmd.append( 1, 0x2C );
           #else
            // Clock len data bytes in on +ve clock edge LSB first
            cmd.append( 1, 0x28 );
           #endif

          #else   // MSB first

           #ifdef SAMPLE_FALLING_EDGE
            // Clock len data bytes in on -ve clock edge MSB first
            cmd.append( 1, 0x24 );
           #else
            // Clock len data bytes in on +ve clock edge MSB first
            cmd.append( 1, 0x20 );
           #endif

          #endif

            cmd.append( 1, (len - 1) & 0xFF );
            cmd.append( 1, (len - 1) >> 8 );

            unsigned reset_attempts = 0;

            try {
                WriteCommand( cmd );
                goto ok;
            }
            BB_CATCH_ALL( 0, stringprintf("BitBabbler::read( %zu ) exception", len).c_str() )

            while( ++reset_attempts < FTDI_INIT_RETRIES ) {
                // We shouldn't ever get here in normal operation, but if for some
                // reason things back up and the USB write request fails, try once
                // to reset the device and reissue the request before bailing out.
                //
                // We could also be here if the read repeatedly returns no data,
                // which could happen if the command written was somehow corrupted
                // in which case resetting the device is also the only way we can
                // be sure of what state it is in.  We bound the maximum number of
                // times that may be attempted while no error is being reported by
                // any of these operations, but any real or permanent error should
                // normally result in bailing out before that limit is reached.
                LogMsg<1>( "BitBabbler::read( %zu ): attempting to reset device", len );
                init_device();
                WriteCommand( cmd );

            ok:
                LogMsg<6>( "BitBabbler::read( %zu ): wrote request", len );

                size_t  count = 0;
                size_t  n = 0;

                do {
                    int ret = ftdi_read_data( &m_ftdi, buf, len - count );

                    if( ret < 0 )
                        ThrowError( _("BitBabbler::read( %zu ) failed"), len );

                    if( ret > 0 )
                    {
                        LogMsg<6>( "BitBabbler::read( %zu ): read %d (n = %zu)",
                                                                    len, ret, n );
                        count += ret;

                        if( count == len )
                        {
                            // This is just to create buffer bloat errors,
                            // mostly for testing the purge recovery code.
                            //WriteCommand( cmd );

                            return len;
                        }

                        n = 0;
                    }

                } while( ++n < FTDI_READ_RETRIES );

                LogMsg<1>( _("BitBabbler::read( %zu ) failed (n = %zu)"), len, n );
            }

            throw Error( _("BitBabbler::read( %zu ) failed after %u reset attempts"),
                                                               len, reset_attempts );
        } //}}}

    }; //}}}


    class Pool : public RefCounted
    { //{{{
    public:

        class Group : public RefCounted
        { //{{{
        public:

            typedef unsigned                                ID;
            typedef uint32_t                                Mask;
            typedef RefPtr< Group >                         Handle;
            typedef std::tr1::unordered_map< ID, Handle >   Map;


            struct Options
            { //{{{

                typedef std::list< Options >    List;

                Group::ID   groupid;
                size_t      size;


                Options( const char *arg )
                {
                    char           *e;
                    unsigned long   v = strtoul( arg, &e, 10 );

                    if( e == arg || *e != ':' || v > unsigned(-1) )
                        throw Error( _("Invalid --group-size option '%s'"), arg );

                    groupid = v;
                    size    = StrToScaledUL( e + 1, 1024 );
                }

            }; //}}}


        private:

            Pool               *m_pool;
            ID                  m_id;
            size_t              m_size;
            uint8_t            *m_buf;
            Mask                m_filled;
            Mask                m_mask;
            unsigned            m_members;
            pthread_mutex_t     m_mutex;


        public:

            Group( Pool *p, ID group_id, size_t size )
                : m_pool( p )
                , m_id( group_id )
                , m_size( powof2_up(size) )
                , m_buf( new uint8_t[m_size] )
                , m_filled( 0 )
                , m_mask( 0 )
                , m_members( 0 )
            {
                Log<2>( "+ Pool::Group( %u, %zu )\n", m_id, m_size );

                pthread_mutex_init( &m_mutex, NULL );
            }

            ~Group()
            {
                Log<2>( "- Pool::Group( %u, %zu )\n", m_id, m_size );

                pthread_mutex_destroy( &m_mutex );
                delete [] m_buf;
            }


            ID GetID() const
            {
                return m_id;
            }

            size_t GetSize() const
            {
                return m_size;
            }


            Mask GetNextMask()
            { //{{{

                ScopedMutex     lock( &m_mutex );

                if( m_id == 0 )
                    return 0;

                for( Mask i = 1; i; i <<= 1 )
                {
                    if( (m_mask & i) == 0 )
                    {
                        m_mask |= i;
                        m_members++;
                        return i;
                    }
                }

                throw Error( _("Pool::Group %u is full"), m_id );

            } //}}}

            void ReleaseMask( Mask i )
            { //{{{

                ScopedMutex     lock( &m_mutex );

                if( m_id == 0 )
                    return;

                if( (m_mask & i) == 0 )
                {
                    // This shouldn't ever happen, but we don't want to
                    // corrupt the m_members count if it somehow does.
                    // We don't want to throw, since we're probably in
                    // a destructor (or similar unwinding) somewhere.
                    Log<0>( _("Pool::Group(%u)::ReleaseMask( %x ): "
                              "mask was not allocated (%x)\n"), m_id, i, m_mask );
                    return;
                }

                m_mask &= ~i;
                m_filled &= ~i;
                m_members--;

            } //}}}


            void AddEntropy( Mask m, uint8_t *b, size_t len )
            { //{{{

                if( len == 0 )
                    return;

                // This should never happen
                if( len != m_size )
                    throw Error( _("Pool::Group(%u:%x)::AddEntropy: len %zu != group size %zu"),
                                                                        m_id, m, len, m_size );

                ScopedMutex     lock( &m_mutex );

                if( m_id == 0 || m_members == 1 )
                {
                    // short-circuit directly to the main pool if there is
                    // only one source in this group (or if this is group 0).
                    m_filled = 0;

                    lock.Unlock();
                    m_pool->AddEntropy( b, len );
                    return;
                }

                if( ! m_filled )
                {
                    memcpy( m_buf, b, len );
                    m_filled = m;

                } else {

                    for( size_t i = 0; i < len; ++i )
                        m_buf[i] ^= b[i];

                    m_filled |= m;
                }

                Log<5>("Group %u:%x: filled %x\n", m_id, m, m_filled);

                if( m_filled == m_mask )
                {
                    uint8_t buf[m_size];

                    memcpy( buf, m_buf, m_size );
                    m_filled = 0;

                    lock.Unlock();
                    m_pool->AddEntropy( buf, m_size );
                }

            } //}}}

        }; //}}}


    private:

        struct Source : public RefCounted
        { //{{{

            typedef RefPtr< Source >        Handle;
            typedef std::list< Handle >     List;


            Pool                   *pool;
            uint8_t                *buf;
            size_t                  size;
            Group::Handle           group;
            Group::Mask             groupmask;
            BitBabbler::Handle      babbler;
            pthread_t               thread;


            Source( Pool                       *p,
                    const Group::Handle        &g,
                    const BitBabbler::Handle   &b )
                : pool( p )
                , size( g->GetSize() * (1u << b->GetFolding()) )
                , group( g )
                , groupmask( g->GetNextMask() )
                , babbler( b )
            {
                Log<2>( "+ Pool::Source( %u:%u, %zu, %s )\n", group->GetID(), groupmask,
                                                    size, babbler->GetSerial().c_str() );
                if( size < babbler->GetChunkSize() )
                    throw Error( _("Pool::Source( %u:%u, %s ): size %zu < chunksize %u"),
                                   group->GetID(), groupmask, babbler->GetSerial().c_str(),
                                                        size, babbler->GetChunkSize() );

                buf = new uint8_t[size];

                int ret = pthread_create( &thread, NULL, Pool::source_thread, this );

                if( ret )
                {
                    group->ReleaseMask( groupmask );
                    delete [] buf;

                    throw SystemError( ret, _("Pool::Source: failed to create thread") );
                }
            }

            ~Source()
            {
                Log<2>( "- Pool::Source( %u:%u, %zu, %s )\n", group->GetID(), groupmask,
                                                    size, babbler->GetSerial().c_str() );
                group->ReleaseMask( groupmask );
                delete [] buf;
            }

        }; //}}}


        struct WriteFD : public RefCounted
        {  //{{{

            typedef RefPtr< WriteFD >   Handle;
            typedef void (*Completion)(void *user_data);

            Pool           *pool;
            int             fd;
            size_t          len;
            Completion      completion_handler;
            void           *user_data;

            WriteFD( Pool *p, int fd, size_t len, Completion handler = NULL,
                                                     void *user_data = NULL )
                : pool( p )
                , fd( fd )
                , len( len )
                , completion_handler( handler )
                , user_data( user_data )
            {}

        }; //}}}


        typedef std::list< pthread_t >      ThreadList;


        uint8_t            *m_buf;
        size_t              m_size;
        size_t              m_fill;
        size_t              m_next;

        std::string         m_kerneldev;

        Group::Map          m_groups;
        Source::List        m_sources;
        ThreadList          m_threads;

        pthread_mutex_t     m_mutex;
        pthread_cond_t      m_sourcecond;
        pthread_cond_t      m_sinkcond;



        bool PoolIsFull()
        {
            ScopedMutex     lock( &m_mutex );
            return m_fill == m_size;
        }

        void AddEntropy( uint8_t *buf, size_t len )
        { //{{{

            if( len == 0 )
                return;

            ScopedMutex     lock( &m_mutex );
            size_t          n = 0;

            if( m_fill < m_size )
            {
                size_t  b = std::min( m_size - m_fill, len );

                Log<5>( "Pool::AddEntropy: add %zu / %zu octets at %zu / %zu\n",
                                                        b, len, m_fill, m_size );
                memcpy( m_buf + m_fill, buf, b );

                n = b;
                m_fill += b;

                pthread_cond_broadcast( &m_sinkcond );
            }

            while( n < len )
            {
                size_t  b = std::min( m_size - m_next, len - n );

                Log<5>( "Pool::AddEntropy: mix %zu / %zu octets at %zu / %zu\n",
                                                        b, len, m_next, m_size );
                for( size_t i = 0; i < b; ++i )
                    m_buf[m_next + i] ^= buf[n + i];

                n += b;
                m_next += b;

                if( m_next >= m_size )
                    m_next = 0;
            }

        } //}}}


        void detach_source( const Source::Handle &s )
        { //{{{

            ScopedMutex     lock( &m_mutex );

            for( Source::List::iterator i = m_sources.begin(),
                                        e = m_sources.end(); i != e; ++ i )
            {
                if( *i == s )
                {
                    m_sources.erase( i );
                    pthread_detach( s->thread );
                    return;
                }
            }

        } //}}}

        BB_NORETURN
        void do_source_thread( const Source::Handle &s )
        { //{{{

            // All times in milliseconds
            static const unsigned   INITIAL_SLEEP   = 100;
            static const unsigned   MIN_SLEEP       = 500;      // really 800
            static const unsigned   MAX_SLEEP       = 60000;    // really 102400

            s->babbler->LogMsg<3>( "Pool: begin source_thread" );

            // At rates of 5Mbps or greater, wait for the first Ent8 test results
            // before declaring the source is generating an acceptable quality of
            // entropy.  Below that let it come online if the FIPS tests aren't
            // rejecting it, with at least 20 consecutive blocks having passed.
            HealthMonitor   qa( s->babbler->GetSerial(),
                                s->babbler->GetBitrate() < 5000000 );

            size_t          read_size   = s->babbler->GetChunkSize();
            unsigned        fold        = s->babbler->GetFolding();
            bool            no_qa       = s->babbler->NoQA();
            unsigned        sleep_for   = 0;
            int             oldstate;

            for(;;)
            {
                if( sleep_for > MIN_SLEEP )
                {
                    timespec        wait_until;

                    GetFutureTimespec( wait_until, sleep_for );

                    ScopedMutex     lock( &m_mutex );

                    int ret = pthread_cond_timedwait( &m_sourcecond, &m_mutex, &wait_until );

                    if( ret && ret != ETIMEDOUT )
                        throw SystemError( ret, "pthread_cond_timedwait failed: %s",
                                                                    strerror(ret) );
                }


                for( size_t p = 0, n = 0; p <= s->size - read_size; p += n )
                {
                    pthread_testcancel();
                    pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate );

                    n = s->babbler->read( s->buf + p, read_size );

                    pthread_setcancelstate( oldstate, NULL );
                }

                size_t n = FoldBytes( s->buf, s->size, fold );


                if( PoolIsFull() )
                {
                    // If the pool is already (still) full, start to throttle back
                    // on how often we keep mixing more new entropy into it.  It's
                    // not that it hurts to do that, but there's probably not much
                    // point burning CPU cycles to do so as fast as possible while
                    // nobody is actually consuming what we alredy have.
                    if( sleep_for == 0 )
                        sleep_for = INITIAL_SLEEP;

                    else if( sleep_for < MAX_SLEEP )
                        sleep_for *= 2;
                }
                else
                    sleep_for = 0;

                // Don't idle if the QA check failed.  We want to find out as quickly
                // as possible if that was just a transient spike outside the limits
                // (which being random is always possible, however rare it may be),
                // and bring the device back on line if that's what it appears to be.
                // If it stays bad, then the sysadmin is going to need to take some
                // action of their own in response to the alert, and this likewise
                // will ensure they have as much data as possible, as quickly as
                // possible to base that decision on.
                if( qa.Check( s->buf, n ) || no_qa )
                    s->group->AddEntropy( s->groupmask, s->buf, n );
                else
                    sleep_for = 0;
            }

        } //}}}

        static void *source_thread( void *p )
        { //{{{

            Source::Handle  s = static_cast<Source*>( p );

            try {
                s->pool->do_source_thread( s );
            }
            catch( const abi::__forced_unwind& )
            {
                s->babbler->LogMsg<3>( "Pool: source_thread cancelled" );
                throw;
            }
            BB_CATCH_STD( 0, s->babbler->MsgStr( _("uncaught source_thread exception") ).c_str() )

            s->pool->detach_source( s );
            return NULL;

        } //}}}


        void detach_thread( pthread_t p )
        { //{{{

            ScopedMutex     lock( &m_mutex );

            for( ThreadList::iterator i = m_threads.begin(),
                                      e = m_threads.end(); i != e; ++i )
            {
                if( *i == p )
                {
                    m_threads.erase( i );
                    pthread_detach( p );
                    return;
                }
            }

        } //}}}

        static void *writefd_thread( void *p )
        { //{{{

            WriteFD::Handle     w = static_cast<WriteFD*>( p );

            try {
                try {
                    w->pool->WriteToFD( w->fd, w->len );

                    Log<3>( "Pool: writefd_thread completed\n" );
                }
                catch( const abi::__forced_unwind& )
                {
                    Log<3>( "Pool: writefd_thread cancelled\n" );
                    throw;
                }
                BB_CATCH_STD( 0, _("uncaught writefd_thread exception") )

                if( w->completion_handler )
                    w->completion_handler( w->user_data );
            }
            catch( const abi::__forced_unwind& )
            {
                Log<3>( "Pool: writefd_thread cancelled\n" );
                throw;
            }
            BB_CATCH_STD( 0, _("uncaught writefd_thread completion exception") )

            w->pool->detach_thread( pthread_self() );
            return NULL;

        } //}}}

        static void *feedkernel_thread( void *p )
        { //{{{

            Pool    *pool = static_cast<Pool*>( p );

            try {
                pool->FeedKernelEntropy( pool->m_kerneldev );
            }
            catch( const abi::__forced_unwind& )
            {
                Log<3>( "Pool: feedkernel_thread cancelled\n" );
                throw;
            }
            BB_CATCH_STD( 0, _("uncaught feedkernel_thread exception") )

            pool->detach_thread( pthread_self() );
            return NULL;

        } //}}}


    public:

        typedef RefPtr< Pool >      Handle;


        Pool( size_t size = 65536 )
            : m_size( size )
            , m_fill( 0 )
            , m_next( 0 )
        { //{{{

            Log<2>( "+ Pool( %zu )\n", m_size );

            m_buf = new uint8_t[size];

            pthread_mutex_init( &m_mutex, NULL );
            pthread_cond_init( &m_sourcecond, NULL );
            pthread_cond_init( &m_sinkcond, NULL );

        } //}}}

        ~Pool()
        { //{{{

            pthread_mutex_lock( &m_mutex );

            Log<3>( "Pool: terminating threads\n" );
            for( ThreadList::iterator i = m_threads.begin(),
                                      e = m_threads.end(); i != e; ++i )
                pthread_cancel( *i );

            Log<3>( "Pool: terminating sources\n" );
            for( Source::List::iterator i = m_sources.begin(),
                                        e = m_sources.end(); i != e; ++ i )
                pthread_cancel( (*i)->thread );


            Log<3>( "Pool: waiting for thread termination\n" );
            while( ! m_threads.empty() )
            {
                pthread_t   p = m_threads.back();

                m_threads.pop_back();

                pthread_mutex_unlock( &m_mutex );
                pthread_join( p, NULL );
                pthread_mutex_lock( &m_mutex );
            }

            Log<3>( "Pool: waiting for source termination\n" );
            while( ! m_sources.empty() )
            {
                pthread_t   p = m_sources.back()->thread;

                m_sources.pop_back();

                pthread_mutex_unlock( &m_mutex );
                pthread_join( p, NULL );
                pthread_mutex_lock( &m_mutex );
            }

            pthread_mutex_unlock( &m_mutex );

            pthread_cond_destroy( &m_sinkcond );
            pthread_cond_destroy( &m_sourcecond );
            pthread_mutex_destroy( &m_mutex );

            delete [] m_buf;

            Log<2>( "- Pool( %zu )\n", m_size );

        } //}}}


        // Group size will be rounded up to a power of 2
        void AddGroup( Group::ID group_id, size_t size )
        { //{{{

            Log<2>( "Pool::AddGroup( %u, %zu )\n", group_id, size );

            ScopedMutex             lock( &m_mutex );
            Group::Map::iterator    i = m_groups.find( group_id );

            if( i != m_groups.end() )
                throw Error( _("Pool::AddGroup( %u, %zu ): group already exists"),
                                                                group_id, size );

            m_groups[group_id] = new Group( this, group_id, size );

        } //}}}

        void AddSource( Group::ID group_id, const BitBabbler::Handle &babbler )
        { //{{{

            babbler->LogMsg<2>( "Pool::AddSource: adding to group %u", group_id );

            ScopedMutex             lock( &m_mutex );
            Group::Map::iterator    gi = m_groups.find( group_id );
            Group::Handle           g;

            if( gi == m_groups.end() )
            {
                g = new Group( this, group_id, m_size );
                m_groups[group_id] = g;

            } else {

                g = gi->second;
            }

            m_sources.push_back( new Source( this, g, babbler ) );

        } //}}}

        void RemoveSource( const std::string &serial )
        { //{{{

            ScopedMutex     lock( &m_mutex );

            for( Source::List::iterator i = m_sources.begin(),
                                        e = m_sources.end(); i != e; ++ i )
            {
                if( (*i)->babbler->GetSerial() == serial )
                {
                    pthread_t   p = (*i)->thread;
                    m_sources.erase( i );

                    lock.Unlock();
                    pthread_cancel( p );
                    pthread_join( p, NULL );
                    return;
                }
            }

        } //}}}


        // Will block until it can return min(len,poolsize) octets
        size_t read( uint8_t *buf, size_t len )
        { //{{{

            Log<5>( "Pool::read( %zu )\n", len );

            ScopedMutex     lock( &m_mutex );

            while( m_fill < m_size && m_fill < len )
                pthread_cond_wait( &m_sinkcond, &m_mutex );

            size_t  n = std::min( m_fill, len );

            memcpy( buf, m_buf + (m_fill - n), n );
            m_fill -= n;

            pthread_cond_broadcast( &m_sourcecond );

            Log<5>( "Pool::read( %zu ) returning %zu\n", len, n );
            return n;

        } //}}}


        void WriteToFD( int fd, size_t len = 0 )
        { //{{{

            uint8_t     buf[65536];

            for(;;)
            {
                size_t      b = len ? std::min( len, sizeof(buf) ) : sizeof(buf);
                size_t      n = read( buf, b );

                for( size_t c = n; c; )
                {
                    ssize_t w = write( fd, buf + n - c, c );

                    if( w < 0 )
                        throw SystemError( _("Pool::WriteToFD( %d ) failed"), fd );

                    if( w == 0 )
                        throw Error( _("Pool::WriteToFD( %d ) EOF"), fd );

                    c -= w;
                }

                if( len && (len -= n) == 0 )
                    return;
            }

        } //}}}

        void WriteToFDAsync( int fd, size_t len = 0, WriteFD::Completion handler = NULL,
                                                                 void *user_data = NULL )
        { //{{{

            ScopedMutex     lock( &m_mutex );
            WriteFD        *w = new WriteFD( this, fd, len, handler, user_data );
            pthread_t       p;

            int ret = pthread_create( &p, NULL, writefd_thread, w );

            if( ret )
            {
                delete w;
                throw SystemError( ret, _("Pool::WriteToFDAsync( %d, %zu ): "
                                          "failed to create thread"), fd, len );
            }

            m_threads.push_back( p );

        } //}}}


        BB_NORETURN
        void FeedKernelEntropy( const std::string &dev = "/dev/random" )
        { //{{{

           #if EM_PLATFORM_LINUX

            // Each time we wake up, either because the kernel pool has fallen
            // below its watermark or due to the timer expiring, we read a full
            // FIPS 140-2 analysis block (20k bits) from our own pool, and if
            // it passes the QA checks we then fold it twice to give us 5k bits
            // to pass to the kernel.  We then pass it through the QA tests yet
            // again, partly to be completely paranoid, and partly so that we
            // can query the QA statistics at that stage too, and if it still
            // passes, we give it to the kernel.
            //
            // If all is working well, the QA testing feeding our pool should
            // ensure that it has very near to 8 bits of entropy per byte, so
            // for each iteration here we take ~20k bits of entropy, fold it
            // to 5k bits for the kernel to process, of which with its default
            // pool size it will credit us for at most 4096 bits of entropy.
            // That should mean the estimate of 8 bits of entropy per byte that
            // we tell it we are providing should be reasonable even if we are
            // a hair below that in the worst case.

            int fd = open( dev.c_str(), O_RDWR );

            if( fd < 0 )
                throw SystemError( _("Pool::FeedKernelEntropy: failed to open %s"),
                                                                    dev.c_str() );

            const unsigned  N       = QA::FIPS::BUFFER_SIZE; // 20kbits for FIPS test.
            const unsigned  folds   = 2;
            const int       timeout = 60000;                 // 60sec

            union {
                uint8_t                 b[N + sizeof(struct rand_pool_info)];
                struct rand_pool_info   rpi;
            };

            HealthMonitor   qa( "Pool" );
            uint8_t        *buf = reinterpret_cast<uint8_t*>( rpi.buf );

            HealthMonitor   qa2( "Kernel" );
            uint8_t         b2[N];
            size_t          b2_fill = 0;

            bool            source_ok;
            bool            folded_ok = false;

            for(;;)
            {
                struct pollfd   p = { fd: fd, events: POLLOUT, revents: 0 };
                int             r = poll( &p, 1, timeout );
                size_t          n;

                if( r < 0 )
                    throw SystemError( _("Pool::FeedKernelEntropy: poll failed") );

                do {
                    n = read( buf, N );

                    if( ! (source_ok = qa.Check( buf, n )) )
                    {
                        b2_fill = 0;
                        continue;
                    }

                    n = FoldBytes( buf, n, folds );

                    memcpy( b2 + b2_fill, buf, n );
                    b2_fill += n;

                    if( b2_fill >= N )
                    {
                        b2_fill   = 0;
                        folded_ok = qa2.Check( b2, N );
                    }

                } while( ! source_ok || ! folded_ok );


                rpi.entropy_count = n * 8;
                rpi.buf_size = n;

                if( ioctl( fd, RNDADDENTROPY, &rpi ) )
                    throw SystemError( _("Pool::FeedKernelEntropy: ioctl failed") );
            }

           #else
            (void)dev;
            throw Error("Kernel entropy not supported on this platform");
           #endif

        } //}}}

        void FeedKernelEntropyAsync( const std::string &dev = "/dev/random" )
        { //{{{

            ScopedMutex     lock( &m_mutex );
            pthread_t       p;

            m_kerneldev = dev;

            int ret = pthread_create( &p, NULL, feedkernel_thread, this );

            if( ret )
                throw SystemError( ret, _("Pool::FeedKernelEntropyAsync( %s ): "
                                          "failed to create thread"), dev.c_str() );

            m_threads.push_back( p );

        } //}}}

    }; //}}}


    class DevList
    { //{{{
    public:

        DevList() {}

        virtual ~DevList() {}


        virtual void AddDevicesToPool( const Pool::Handle               &pool,
                                       const BitBabbler::Options        &default_options,
                                       const BitBabbler::Options::List  &device_options ) = 0;


        virtual unsigned GetNumDevices() const = 0;

        // Get a device by index number
        virtual struct usb_device *GetDevice( unsigned idx ) const = 0;

        // Get a device by serial ID
        virtual struct usb_device *GetDevice( const std::string &serial ) const = 0;

    }; //}}}


   #if HAVE_LIBUDEV

    class UDEVMonitor : public DevList
    { //{{{
    protected:

        struct Device
        { //{{{
        private:

            void get_usb_device( unsigned vendorid, unsigned productid )
            { //{{{

                if( usb_find_busses() < 0 )
                    throw Error( _("UDEVMonitor: usb_find_busses failed") );

                if( usb_find_devices() < 0 )
                    throw Error( _("UDEVMonitor: usb_find_devices failed") );

                for( struct usb_bus *bus = usb_get_busses(); bus; bus = bus->next )
                {
                    // In theory we could use bus->location here, but setting that
                    // (from dirname) appears to be a Debian specific patch right now.
                    unsigned    bnum = StrToUL( bus->dirname, 10 );

                    if( bnum != busnum )
                        continue;

                    for( struct usb_device *dev = bus->devices; dev; dev = dev->next )
                    {
                        // And in theory we could use dev->devnum here, except it appears
                        // to get set to 0 if the user running this doesn't have permission
                        // to actually access the device.  Which is something of a separate
                        // problem to what we care about here.  It should be ok to let an
                        // unprivileged user scan for what devices are present, that's not
                        // a secret they can't find out in other ways.
                        unsigned    dnum = StrToUL( dev->filename, 10 );

                        if( dnum != devnum )
                            continue;

                        if( dev->descriptor.idVendor != vendorid
                         || dev->descriptor.idProduct != productid )
                        {
                            // This really shouldn't ever happen ...
                            Log<0>( _("UDEVMonitor::Device: matched busnum %u and devnum %u "
                                      "but mismatched vendor:product, %u:%u != %u:%u\n"),
                                      busnum, devnum, vendorid, productid, dev->descriptor.idVendor,
                                                                           dev->descriptor.idProduct );
                            continue;
                        }

                        USBInfo     info( dev );

                        // This shouldn't normally ever happen, but it can if the EEPROM was
                        // rewritten or corrupted in some way.  Warn because probably the
                        // device just needs to be re-enumerated to correct the kernel's idea
                        // of what is really there.  It can also happen if the user calling
                        // this does not have permission to actually read from the device.
                        if( mfg != info.mfg || product != info.product || serial != info.serial )
                            Log<0>( _("UDEVMonitor::Device: "
                                      "expecting mfg '%s', product '%s', serial '%s', "
                                      "but device %u-%u returned '%s', '%s', '%s'\n"),
                                      mfg.c_str(), product.c_str(), serial.c_str(), busnum, devnum,
                                      info.mfg.c_str(), info.product.c_str(), info.serial.c_str() );

                        usb_device = dev;
                        return;
                    }
                }

                throw Error( _("UDEVMonitor::Device: "
                               "failed to find usb_device( %04x, %04x, '%s' )"),
                                                vendorid, productid, serial.c_str() );
            } //}}}


        public:

            typedef std::list< Device >     List;

            std::string         mfg;
            std::string         product;
            std::string         serial;
            unsigned            busnum;
            unsigned            devnum;
            std::string         devport;        // ATTR{devpath}, not to be confused with
            std::string         devpath;        // this one, the sysfs 'devpath' ...
            struct usb_device  *usb_device;


            Device( unsigned            vendorid,
                    unsigned            productid,
                    const std::string  &mfg,
                    const std::string  &product,
                    const std::string  &serial,
                    const std::string  &busnum,
                    const std::string  &devnum,
                    const std::string  &devport,
                    const std::string  &devpath )
                : mfg( mfg )
                , product( product )
                , serial( serial )
                , busnum( StrToUL(busnum, 10) )
                , devnum( StrToUL(devnum, 10) )
                , devport( devport )
                , devpath( devpath )
            {
                get_usb_device( vendorid, productid );
            }

            Device( const std::string  &vendorid,
                    const std::string  &productid,
                    const std::string  &mfg,
                    const std::string  &product,
                    const std::string  &serial,
                    const std::string  &busnum,
                    const std::string  &devnum,
                    const std::string  &devport,
                    const std::string  &devpath )
                : mfg( mfg )
                , product( product )
                , serial( serial )
                , busnum( StrToUL(busnum, 10) )
                , devnum( StrToUL(devnum, 10) )
                , devport( devport )
                , devpath( devpath )
            {
                get_usb_device( StrToUL(vendorid, 16), StrToUL(productid, 16) );
            }


            DeviceInfo GetDeviceInfo() const
            {
                return DeviceInfo( mfg, product, serial, busnum, devnum, devport );
            }

        }; //}}}


    private:

        static const unsigned       m_qsize = 4096;

        struct udev                *m_udev;
        struct udev_monitor        *m_mon;

        pthread_t                   m_actor_thread;
        pthread_t                   m_monitor_thread;

        pthread_mutex_t             m_action_mutex;
        pthread_cond_t              m_action_cond;
        struct udev_device         *m_deviceq[ m_qsize ];
        unsigned                    m_deviceq_rd;
        unsigned                    m_deviceq_wr;

        std::string                 m_vendorid;
        std::string                 m_productid;

        // You must hold m_device_mutex to access these
        mutable pthread_mutex_t     m_device_mutex;
        Pool::Handle                m_pool;
        BitBabbler::Options         m_default_options;
        BitBabbler::Options::List   m_device_options;
        Device::List                m_devices;


        void dump_udev_data( struct udev_device *d )
        { //{{{

            const char *action      = udev_device_get_action( d );
            const char *sysname     = udev_device_get_sysname( d );
            const char *driver      = udev_device_get_driver( d );
            const char *subsys      = udev_device_get_subsystem( d );
            const char *devtype     = udev_device_get_devtype( d );
            const char *devpath     = udev_device_get_devpath( d );
            const char *devnode     = udev_device_get_devnode( d );
            const char *syspath     = udev_device_get_syspath( d );
            const char *vendorid    = udev_device_get_sysattr_value( d, "idVendor" );
            const char *productid   = udev_device_get_sysattr_value( d, "idProduct" );
            const char *serial      = udev_device_get_sysattr_value( d, "serial" );
         // const char *vendorid    = udev_device_get_property_value( d, "ID_VENDOR_ID" );
         // const char *productid   = udev_device_get_property_value( d, "ID_MODEL_ID" );
         // const char *serial      = udev_device_get_property_value( d, "ID_SERIAL_SHORT" );

            Log<0>( "action: %s, sysname: %s, driver: %s\n", action, sysname, driver );
            Log<0>( "subsys: %s, devtype: %s, devpath: %s\n", subsys, devtype, devpath );
            Log<0>( "devnode: %s, syspath: %s\n", devnode, syspath );
            Log<0>( "vendorid: %s, productid: %s, serial: %s\n", vendorid, productid, serial );


            struct udev_list_entry  *attrlist, *a;

            attrlist = udev_device_get_sysattr_list_entry( d );

            udev_list_entry_foreach( a, attrlist)
            {
                const char *key = udev_list_entry_get_name( a );

                if( ! key )
                    continue;

                const char *val = udev_device_get_sysattr_value( d, key );

                Log<0>( "  attribute %s = '%s'\n", key, val );
            }


            struct udev_list_entry  *proplist, *p;

            proplist = udev_device_get_properties_list_entry( d );

            udev_list_entry_foreach( p, proplist )
            {
                const char *key = udev_list_entry_get_name( p );

                if( ! key )
                    continue;

                const char *val = udev_list_entry_get_value( p );

                Log<0>( "  property  %s = '%s'\n", key, val );
            }


            struct udev_list_entry  *taglist, *t;

            taglist = udev_device_get_tags_list_entry( d );

            udev_list_entry_foreach( t, taglist )
            {
                const char *key = udev_list_entry_get_name( t );

                if( ! key )
                    continue;

                const char *val = udev_list_entry_get_value( t );

                Log<0>( "  tag       %s = '%s'\n", key, val );
            }

        } //}}}


    protected:

        Device::List GetDevices() const
        { //{{{

            ScopedMutex     lock( &m_device_mutex );
            return m_devices;

        } //}}}

        virtual void AddDevice( const Device &d )
        { //{{{

            if( d.serial.empty() || ! m_pool )
                return;

            if( m_device_options.empty() )
            {
                m_pool->AddSource( m_default_options.group,
                                   new BitBabbler( d.usb_device, m_default_options ) );
                return;
            }

            for( BitBabbler::Options::List::iterator i = m_device_options.begin(),
                                                     e = m_device_options.end();
                                                     i != e; ++i )
            {
                if( i->device_id == d.serial )
                {
                    m_pool->AddSource( i->group, new BitBabbler( d.usb_device, *i ) );
                    return;
                }
            }

        } //}}}

        virtual void RemoveDevice( const Device &d )
        { //{{{

            if( ! d.serial.empty() && m_pool != NULL )
                m_pool->RemoveSource( d.serial );

        } //}}}


    private:

        void add_device( const Device &d )
        { //{{{

            ScopedMutex     lock( &m_device_mutex );

            Log<2>( _("UDEVMonitor: add '%s' %s\n"), d.serial.c_str(), d.devpath.c_str() );

            m_devices.push_back( d );

            try {
                AddDevice( d );
            }
            BB_CATCH_ALL( 0, "UDEVMonitor::AddDevice failed" )

        } //}}}

        void remove_device( const char *devpath )
        { //{{{

            ScopedMutex     lock( &m_device_mutex );


            for( Device::List::iterator i = m_devices.begin(),
                                        e = m_devices.end(); i != e; ++i )
            {
                if( i->devpath == devpath )
                {
                    try {
                        Log<2>( _("UDEVMonitor: remove %s: %s\n"),
                                i->GetDeviceInfo().VerboseStr().c_str(), devpath );

                        RemoveDevice( *i );
                    }
                    BB_CATCH_ALL( 0, "UDEVMonitor::RemoveDevice failed" )

                    m_devices.erase( i );
                    return;
                }
            }

            Log<4>( _("UDEVMonitor: remove %s\n"), devpath );

        } //}}}


        // Caller must hold m_device_mutex
        struct usb_device *get_device( unsigned idx ) const
        { //{{{

            unsigned    n = idx;

            for( Device::List::const_iterator i = m_devices.begin(),
                                              e = m_devices.end(); i != e; ++i, --n )
                if( n == 0 )
                    return i->usb_device;

            throw Error( _("UDEVMonitor::GetDevice( %u ): out of range"), idx);

        } //}}}

        // Caller must hold m_device_mutex
        struct usb_device *get_device( const std::string &serial ) const
        { //{{{

            for( Device::List::const_iterator i = m_devices.begin(),
                                              e = m_devices.end(); i != e; ++i )
                if( i->serial == serial )
                    return i->usb_device;

            throw Error( _("UDEVMonitor::GetDevice( %s ): no such device"),
                                                            serial.c_str() );
        } //}}}


        // Run the queue of device notifications.
        BB_NORETURN
        void __actor_thread()
        { //{{{

            int oldstate;

            for(;;)
            {
                while( m_deviceq_wr - m_deviceq_rd > 0 )
                {
                    // There is a window here, where if we get cancelled, a few
                    // new device notifications may have already been queued.
                    // We could install a fancier cleanup handler, to catch any
                    // that we might otherwise leak - but since right now, this
                    // thread should only be exiting if the whole app is being
                    // torn down, we'll just let the Big Reaper get them for us.

                    pthread_testcancel();
                    pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate );

                    struct udev_device *d = m_deviceq[m_deviceq_rd++ % m_qsize];

                    if( opt_verbose > 3 )
                        dump_udev_data( d );


                    const char *action  = udev_device_get_action( d );
                    const char *devpath = udev_device_get_devpath( d );

                    if( ! action || ! devpath )
                    {
                        // Ensure strcmp won't explode if udev is braindead
                        Log<0>( _("UDEVMonitor: event with no action or devpath\n") );
                    }
                    else if( strcmp(action, "remove") == 0 )
                    {
                        remove_device( devpath );
                    }
                    else if( strcmp(action, "add") == 0 )
                    {
                      // Alternatively, but less reliably:
                      //const char *vendorid    = udev_device_get_property_value( d, "ID_VENDOR_ID" );
                      //const char *productid   = udev_device_get_property_value( d, "ID_MODEL_ID" );
                        const char *vendorid    = udev_device_get_sysattr_value( d, "idVendor" );
                        const char *productid   = udev_device_get_sysattr_value( d, "idProduct" );

                        if( vendorid && productid && vendorid == m_vendorid && productid == m_productid )
                        {
                          //const char *serial  = udev_device_get_property_value( d, "ID_SERIAL_SHORT" );
                            const char *serial  = udev_device_get_sysattr_value( d, "serial" );

                            try {
                              //const char *mfg     = udev_device_get_property_value( d, "ID_VENDOR" );
                              //const char *product = udev_device_get_property_value( d, "ID_MODEL" );
                                const char *mfg     = udev_device_get_sysattr_value( d, "manufacturer" );
                                const char *product = udev_device_get_sysattr_value( d, "product" );
                                const char *busnum  = udev_device_get_sysattr_value( d, "busnum" );
                                const char *devnum  = udev_device_get_sysattr_value( d, "devnum" );
                                const char *devport = udev_device_get_sysattr_value( d, "devpath" );

                                add_device( Device( m_vendorid, m_productid,
                                                    mfg     ? mfg     : "",
                                                    product ? product : "",
                                                    serial  ? serial  : "",
                                                    busnum  ? busnum  : "",
                                                    devnum  ? devnum  : "",
                                                    devport ? devport : "",
                                                    devpath ) );
                            }
                            BB_CATCH_ALL( 0, _("UDEVMonitor: add event exception") )
                        }
                    }

                    udev_device_unref( d );
                    pthread_setcancelstate( oldstate, NULL );
                }


                ScopedMutex     lock( &m_action_mutex );

                pthread_cond_wait( &m_action_cond, &m_action_mutex );
            }

        } //}}}

        // Wait for udev to chirp at us.
        BB_NORETURN
        void __monitor_thread()
        { //{{{

            for(;;)
            {
                struct udev_device *d = udev_monitor_receive_device( m_mon );
                int                 oldstate;

                if( ! d )
                    continue;

                pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate );

                if( m_deviceq_wr - m_deviceq_rd < m_qsize )
                {
                    m_deviceq[m_deviceq_wr % m_qsize] = d;
                    m_deviceq_wr++;

                    ScopedMutex     lock( &m_action_mutex );

                    pthread_cond_broadcast( &m_action_cond );

                } else {

                    Log<0>( _("UDEVMonitor: *** queue full, packet dropped ***\n") );
                    udev_device_unref( d );
                }
                pthread_setcancelstate( oldstate, NULL );
            }

        } //}}}

        static void *actor_thread( void *p )
        { //{{{

            try {
                static_cast<UDEVMonitor*>(p)->__actor_thread();
            }
            catch( const abi::__forced_unwind& )
            {
                Log<3>( "UDEVMonitor: actor_thread cancelled\n" );
                throw;
            }
            BB_CATCH_STD( 0, _("uncaught actor_thread exception") )

            return NULL;

        } //}}}

        static void *monitor_thread( void *p )
        { //{{{

            try {
                static_cast<UDEVMonitor*>(p)->__monitor_thread();
            }
            catch( const abi::__forced_unwind& )
            {
                Log<3>( "UDEVMonitor: monitor_thread cancelled\n" );
                throw;
            }
            BB_CATCH_STD( 0, _("uncaught monitor_thread exception") )

            return NULL;

        } //}}}


    public:

        UDEVMonitor( unsigned vendorid = BB_VENDOR_ID, unsigned productid = BB_PRODUCT_ID )
            : m_udev( udev_new() )
            , m_mon( NULL )
            , m_deviceq_rd( 0 )
            , m_deviceq_wr( 0 )
            , m_vendorid( stringprintf("%04x", vendorid) )
            , m_productid( stringprintf("%04x", productid) )
        { //{{{

            Log<2>( "+ UDEVMonitor( %04x, %04x )\n", vendorid, productid );

            if( ! m_udev )
                throw Error( _("UDEVMonitor: failed to create udev context") );

            pthread_mutex_init( &m_device_mutex, NULL );

            try {

                usb_init();

                if( usb_find_busses() < 0 )
                    throw Error( _("UDEVMonitor: usb_find_busses failed") );

                if( usb_find_devices() < 0 )
                    throw Error( _("UDEVMonitor: usb_find_devices failed") );


                m_mon = udev_monitor_new_from_netlink( m_udev, "udev" );
                if( ! m_mon )
                    throw Error( _("UDEVMonitor: failed to create udev monitor") );


                // Let the monitor socket block, it has its own thread here
                int     monfd       = udev_monitor_get_fd( m_mon );
                int     socketflags = fcntl( monfd, F_GETFL );

                if( socketflags == -1 )
                    throw SystemError( _("UDEVMonitor: failed to get monitor fd flags") );

                if( fcntl( monfd, F_SETFL, socketflags & ~O_NONBLOCK ) == -1 )
                    throw SystemError( _("UDEVMonitor: failed to clear monitor fd O_NONBLOCK") );


                udev_monitor_filter_add_match_subsystem_devtype( m_mon, "usb", "usb_device" );

                if( udev_monitor_enable_receiving( m_mon ) )
                    throw Error( _("UDEVMonitor: failed to bind udev monitor") );


                struct udev_enumerate   *e = udev_enumerate_new( m_udev );
                struct udev_list_entry  *devlist, *d;

                if( ! e )
                    throw Error( _("UDEVMonitor: failed to create enum context") );

                udev_enumerate_add_match_sysattr( e, "idVendor", m_vendorid.c_str() );
                udev_enumerate_add_match_sysattr( e, "idProduct", m_productid.c_str() );
                udev_enumerate_scan_devices( e );
                devlist = udev_enumerate_get_list_entry( e );

                ScopedMutex     lock( &m_device_mutex );

                udev_list_entry_foreach( d, devlist )
                {
                    const char         *syspath = udev_list_entry_get_name( d );
                    struct udev_device *ud = udev_device_new_from_syspath( m_udev, syspath );

                    if( ! ud )
                    {
                        Log<0>( _("UDEVMonitor: failed to get device from syspath '%s'\n"),
                                                                                syspath );
                        continue;
                    }

                    if( opt_verbose > 3 )
                        dump_udev_data( ud );

                  // Alternatively, but less reliably:
                  //const char *serial  = udev_device_get_property_value( ud, "ID_SERIAL_SHORT" );
                  //const char *mfg     = udev_device_get_property_value( ud, "ID_VENDOR" );
                  //const char *product = udev_device_get_property_value( ud, "ID_MODEL" );
                    const char *serial  = udev_device_get_sysattr_value( ud, "serial" );
                    const char *mfg     = udev_device_get_sysattr_value( ud, "manufacturer" );
                    const char *product = udev_device_get_sysattr_value( ud, "product" );
                    const char *busnum  = udev_device_get_sysattr_value( ud, "busnum" );
                    const char *devnum  = udev_device_get_sysattr_value( ud, "devnum" );
                    const char *devport = udev_device_get_sysattr_value( ud, "devpath" );
                    const char *devpath = udev_device_get_devpath( ud );

                    try {
                        if( devpath )
                        {
                            m_devices.push_back( Device( vendorid, productid,
                                                         mfg     ? mfg     : "",
                                                         product ? product : "",
                                                         serial  ? serial  : "",
                                                         busnum  ? busnum  : "",
                                                         devnum  ? devnum  : "",
                                                         devport ? devport : "",
                                                         devpath ) );
                        } else {
                            // Guard against udev being braindead
                            Log<0>( _("UDEVMonitor: device '%s' with no devpath\n"), syspath );
                        }
                    }
                    catch( ... )
                    {
                        udev_device_unref( ud );
                        udev_enumerate_unref( e );
                        throw;
                    }

                    udev_device_unref( ud );
                }

                lock.Unlock();
                udev_enumerate_unref( e );


                // Now wait in the background for things to change
                pthread_mutex_init( &m_action_mutex, NULL );
                pthread_cond_init( &m_action_cond, NULL );

                int ret = pthread_create( &m_actor_thread, NULL, actor_thread, this );
                if( ret )
                {
                    pthread_cond_destroy( &m_action_cond );
                    pthread_mutex_destroy( &m_action_mutex );

                    throw SystemError( ret, _("UDEVMonitor: failed to create actor thread") );
                }

                ret = pthread_create( &m_monitor_thread, NULL, monitor_thread, this );
                if( ret )
                {
                    pthread_cancel( m_actor_thread );
                    pthread_join( m_actor_thread, NULL );

                    pthread_cond_destroy( &m_action_cond );
                    pthread_mutex_destroy( &m_action_mutex );

                    throw SystemError( ret, _("UDEVMonitor: failed to create monitor thread") );
                }
            }
            catch( ... )
            {
                if( m_mon )
                    udev_monitor_unref( m_mon );

                if( m_udev )
                    udev_unref( m_udev );

                pthread_mutex_destroy( &m_device_mutex );

                throw;
            }

        } //}}}

        virtual ~UDEVMonitor()
        { //{{{

            if( m_mon )
            {
                Log<3>( "UDEVMonitor( %s, %s ): halting monitor threads\n",
                                    m_vendorid.c_str(), m_productid.c_str() );

                pthread_cancel( m_actor_thread );
                pthread_cancel( m_monitor_thread );
                pthread_join( m_actor_thread, NULL );
                pthread_join( m_monitor_thread, NULL );

                udev_monitor_unref( m_mon );
            }

            if( m_udev )
                udev_unref( m_udev );

            pthread_cond_destroy( &m_action_cond );
            pthread_mutex_destroy( &m_action_mutex );
            pthread_mutex_destroy( &m_device_mutex );

            Log<2>( "- UDEVMonitor( %s, %s )\n", m_vendorid.c_str(), m_productid.c_str() );

        } //}}}


        void AddDevicesToPool( const Pool::Handle               &pool,
                               const BitBabbler::Options        &default_options,
                               const BitBabbler::Options::List  &device_options )
        { //{{{

            ScopedMutex     lock( &m_device_mutex );

            m_pool            = pool;
            m_default_options = default_options;
            m_device_options  = device_options;

            if( m_device_options.empty() )
            {
                for( Device::List::const_iterator i = m_devices.begin(),
                                                  e = m_devices.end(); i != e; ++i )
                {
                    if( i->serial.empty() )
                    {
                        // We shouldn't be here in normal use, the user would need to
                        // call this explicitly on a monitor that enumerated bare (or
                        // broken) devices, but don't add them automatically if they do.
                        // If there's some reason to really want to, they can be added
                        // explicitly by index in the loop below.
                        Log<0>( _("UDEVMonitor::AddDevicesToPool: ignoring device %u-%s "
                                  "which has no serial number\n"), i->busnum, i->devport.c_str() );
                        continue;
                    }

                    try {
                        m_pool->AddSource( m_default_options.group,
                                           new BitBabbler( i->usb_device, m_default_options ) );
                    }
                    BB_CATCH_ALL( 0, _("AddDevicesToPool failed") )
                }

            } else {

                for( BitBabbler::Options::List::iterator i = m_device_options.begin(),
                                                         e = m_device_options.end();
                                                         i != e; ++i )
                {
                    // It's not necessarily a hard error for this to fail.
                    // The user may have specified a list of the devices to use
                    // but (all of) those devices may not be plugged in yet.
                    try
                    {
                        struct usb_device   *usb_dev;

                        if( i->device_id.empty() )
                            usb_dev = get_device( i->device_num );
                        else
                            usb_dev = get_device( i->device_id );

                        m_pool->AddSource( i->group, new BitBabbler( usb_dev, *i ) );
                    }
                    BB_CATCH_ALL( 0, _("AddDevicesToPool failed") )
                }
            }

        } //}}}


        const std::string &GetVendorID() const  { return m_vendorid; }
        const std::string &GetProductID() const { return m_productid; }


        unsigned GetNumDevices() const
        {
            ScopedMutex     lock( &m_device_mutex );
            return m_devices.size();
        }

        // Get a device by index number
        struct usb_device *GetDevice( unsigned idx ) const
        {
            ScopedMutex     lock( &m_device_mutex );
            return get_device( idx );
        }

        // Get a device by serial ID
        struct usb_device *GetDevice( const std::string &serial ) const
        {
            ScopedMutex     lock( &m_device_mutex );
            return get_device( serial );
        }

        // Populate a back-insertion sequence with information about all available devices
        template< typename Sequence >
        void GetDeviceInfo( Sequence &info ) const
        { //{{{

            ScopedMutex     lock( &m_device_mutex );

            for( Device::List::const_iterator i = m_devices.begin(),
                                              e = m_devices.end(); i != e; ++i )
                info.push_back( i->GetDeviceInfo() );

        } //}}}

        // List all available devices
        void ListDevices() const
        { //{{{

            ScopedMutex     lock( &m_device_mutex );

            if( m_devices.empty() )
            {
                printf( _("No devices found.\n") );
                return;
            }

            size_t              n = m_devices.size();
            DeviceInfo::Vector  info;

            printf( _P("Have %zu device:\n",
                       "Have %zu devices:\n", n), n );

            n = 0;

            // GetDeviceInfo( info ), but we already hold the mutex.
            // We could just report all this directly in the loop below, but use
            // the standard formatting from DeviceInfo so we only have to maintain
            // that in one place.  We're not going to be time critical here.
            for( Device::List::const_iterator i = m_devices.begin(),
                                              e = m_devices.end(); i != e; ++i )
                info.push_back( i->GetDeviceInfo() );

            for( DeviceInfo::Vector::iterator i = info.begin(),
                                              e = info.end(); i != e; ++i, ++n )
            {
                if( opt_verbose )
                    printf( _("  Device %zu: %s\n"), n, i->VerboseStr().c_str() );
                else
                    printf( _("  Device %zu: %s\n"), n, i->IDStr().c_str() );
            }

        } //}}}

    }; //}}}

    typedef UDEVMonitor     Devices;

   #else   // ! HAVE_LIBUDEV

    class DeviceList : public DevList, public FTDI
    { //{{{
    private:

        struct ftdi_device_list    *m_devs;
        unsigned                    m_ndevs;
        int                         m_vendorid;
        int                         m_productid;


    public:

        DeviceList( int vendorid = BB_VENDOR_ID, int productid = BB_PRODUCT_ID )
            : m_vendorid( vendorid )
            , m_productid( productid )
        {
            Log<2>( "+ DeviceList( %04x, %04x )\n", vendorid, productid );

            int ret = ftdi_usb_find_all( &m_ftdi, &m_devs, vendorid, productid );
            if( ret < 0 )
                ThrowError( _("DeviceList: failed to enumerate devices") );

            m_ndevs = ret;
        }

        ~DeviceList()
        {
            ftdi_list_free( &m_devs );
            Log<2>( "- DeviceList( %04x, %04x )\n", m_vendorid, m_productid );
        }


        void AddDevicesToPool( const Pool::Handle               &pool,
                               const BitBabbler::Options        &default_options,
                               const BitBabbler::Options::List  &device_options )
        { //{{{

            if( device_options.empty() )
            {
                for( struct ftdi_device_list *d = m_devs; d != NULL; d = d->next )
                {
                    if( ! d->dev->descriptor.iSerialNumber )
                    {
                        // We shouldn't be here in normal use, the user would need to
                        // call this explicitly on a DeviceList that enumerated bare (or
                        // broken) devices, but don't add them automatically if they do.
                        // If there's some reason to really want to, they can be added
                        // explicitly by index in the loop below.
                        Log<0>( _("DeviceList::AddDevicesToPool: ignoring device %lu-%u "
                                  "which has no serial number\n"),
                                  StrToUL(d->dev->bus->dirname, 10), d->dev->devnum );
                        continue;
                    }

                    try {
                        pool->AddSource( default_options.group,
                                         new BitBabbler( d->dev, default_options ) );
                    }
                    BB_CATCH_ALL( 0, _("AddDevicesToPool failed") )
                }

            } else {

                for( BitBabbler::Options::List::const_iterator i = device_options.begin(),
                                                               e = device_options.end();
                                                               i != e; ++i )
                {
                    // It's not necessarily a hard error for this to fail.
                    // The user may have specified a list of the devices to use
                    // but (all of) those devices may not be plugged in now.
                    try
                    {
                        struct usb_device   *usb_dev;

                        if( i->device_id.empty() )
                            usb_dev = GetDevice( i->device_num );
                        else
                            usb_dev = GetDevice( i->device_id );

                        pool->AddSource( i->group, new BitBabbler( usb_dev, *i ) );
                    }
                    BB_CATCH_ALL( 0, _("AddDevicesToPool failed") )
                }
            }

        } //}}}


        unsigned GetNumDevices() const
        {
            return m_ndevs;
        }

        // Get a device by index number
        struct usb_device *GetDevice( unsigned idx ) const
        { //{{{

            if( idx >= m_ndevs )
                throw Error( _("DeviceList::GetDevice( %u ): out of range"), idx);

            struct ftdi_device_list *d = m_devs;

            for( unsigned i = 0; i < idx; ++i )
                d = d->next;

            return d->dev;

        } //}}}

        // Get a device by serial ID
        struct usb_device *GetDevice( const std::string &serial ) const
        { //{{{

            for( struct ftdi_device_list *d = m_devs; d != NULL; d = d->next )
            {
                char s[128];

                if( ! d->dev->descriptor.iSerialNumber )
                    continue;

                if( ftdi_usb_get_strings( &m_ftdi, d->dev, NULL, 0, NULL, 0, s, sizeof(s)) )
                {
                    LogError<1>( _("DeviceList::GetDevice( %s ): get strings failed"),
                                                                    serial.c_str() );
                    continue;
                }

                if( serial == s )
                    return d->dev;
            }

            throw Error( _("DeviceList::GetDevice( %s ): no such device"),
                                                         serial.c_str() );
        } //}}}

        // Populate a back-insertion sequence with information about all available devices
        template< typename Sequence >
        void GetDeviceInfo( Sequence &info ) const
        { //{{{

            unsigned n = 0;

            if( m_ndevs == 0 )
                return;

            for( struct ftdi_device_list *d = m_devs; d != NULL; d = d->next, ++n )
            {
                char mfg[128];
                char product[128];
                char serial[128];

                if( ftdi_usb_get_strings( &m_ftdi, d->dev, mfg, sizeof(mfg),
                                                       product, sizeof(product),
                                                        serial, sizeof(serial)) )
                {
                    LogError<1>( _("DeviceList::GetDeviceInfo: failed for device %u"), n );
                    // Push an (almost) empty entry here, so GetDevice( byindex ) will be
                    // correct for the index into the returned vector.  The most probable
                    // reason to be here is the user calling this doesn't have permission
                    // to access the device, in which case d->dev->devnum won't be set
                    // either, so extract that from the filename field here too.
                    info.push_back( DeviceInfo( "", "", "", StrToUL(d->dev->bus->dirname, 10),
                                                            StrToUL(d->dev->filename, 10)) );
                    continue;
                }

                // libftdi doesn't check if these are valid before requesting
                // descriptor strings for them, though it should, so do it here.
                // The serial number in particular will not be set on bare boards.
                if( ! d->dev->descriptor.iManufacturer )
                    mfg[0] = '\0';

                if( ! d->dev->descriptor.iProduct )
                    product[0] = '\0';

                if( ! d->dev->descriptor.iSerialNumber )
                    serial[0] = '\0';

                // In theory we could use d->dev->bus->location here, but setting that
                // (from dirname) appears to be a Debian specific patch right now.
                info.push_back( DeviceInfo( mfg, product, serial,
                                            StrToUL(d->dev->bus->dirname, 10),
                                            d->dev->devnum ) );
            }

        } //}}}

        // List all available devices
        void ListDevices() const
        { //{{{

            if( m_ndevs == 0 )
            {
                printf( _("No devices found.\n") );
                return;
            }

            printf( _P("Have %d device:\n",
                       "Have %d devices:\n", m_ndevs), m_ndevs );


            unsigned            n = 0;
            DeviceInfo::Vector  info;

            GetDeviceInfo( info );

            for( DeviceInfo::Vector::iterator i = info.begin(),
                                              e = info.end(); i != e; ++i, ++n )
            {
                if( opt_verbose )
                    printf( _("  Device %u: %s\n"), n, i->VerboseStr().c_str() );
                else
                    printf( _("  Device %u: %s\n"), n, i->IDStr().c_str() );
            }

        } //}}}

    }; //}}}

    typedef DeviceList      Devices;

   #endif  // HAVE_LIBUDEV


}   // BitB namespace


#endif  // _BB_SECRET_SOURCE_H

// vi:sts=4:sw=4:et:foldmethod=marker
